from typing import List

import pandas as pd
import scipy.io

from graphysio.plotwidgets.curves import CurveItem


def curves_to_matlab(
    curves: List[CurveItem],
    filepath: str,
    index_label: str = "timens",
) -> None:
    sers = [c.series for c in curves]
    data = pd.concat(sers, axis=1).sort_index()
    data["timens"] = data.index.to_numpy()
    scipy.io.savemat(filepath, {"data": data.to_dict("list")})


curves_to_matlab.is_available = True



from typing import List

import pandas as pd

from graphysio.dialogs import askUserValue
from graphysio.plotwidgets import PlotWidget
from graphysio.plotwidgets.curves import CurveItemWithPOI
from graphysio.structures import Parameter, PlotData


def get_perfusion_index(plotwidget: PlotWidget) -> List[CurveItemWithPOI]:
    curvenames = list(plotwidget.curves.keys())
    curvenames = list(plotwidget.curves.keys())
    if len(curvenames) < 1:
        msg = "No curve"
        raise ValueError(msg)
    elif len(curvenames) > 1:
        q = Parameter("Select Curve", curvenames)
        curvename = askUserValue(q)
    else:
        curvename = curvenames[0]

    curve = plotwidget.curves[curvename]
    if (
        "start" not in curve.feetitem.indices
        or curve.feetitem.indices["start"].size < 1
    ):
        msg = "No start information for curve"
        raise ValueError(msg)

    wave = curve.series
    starts = curve.getFeetPoints("start")
    df = pd.concat([wave, starts], axis=1)
    df = df.interpolate(method="index")

    begins, durations = curve.getCycleIndices()
    pivalues = []
    for begin, duration in zip(begins, durations):
        cycle = df.loc[begin : begin + duration]
        auctot = cycle[wave.name].sum()
        aucbase = cycle[starts.name].sum()
        aucpulse = auctot - aucbase
        pi = aucpulse / auctot
        pivalues.append(pi)

    out_series = pd.Series(pivalues, index=begins)
    sname = f"{wave.name}-perf"

    plotdata = PlotData(data=out_series, name=sname)
    return [plotdata]



from typing import List

import pandas as pd

from graphysio.plotwidgets.curves import CurveItem

try:
    import pyarrow.parquet  # noqa: F401
except ImportError:
    is_available = False
else:
    is_available = True


def export_curves(
    curves: List[CurveItem],
    filepath: str,
    index_label: str = "timens",
) -> None:
    sers = [c.series for c in curves]
    data = pd.concat(sers, axis=1).sort_index()
    data.index = pd.to_datetime(data.index, unit="ns", utc=True)
    data.to_parquet(
        filepath,
        engine="pyarrow",
        compression="gzip",
        coerce_timestamps="us",
        allow_truncated_timestamps=True,
    )


export_curves.is_available = is_available



import itertools
import os
import sys
import traceback
from pebble import ProcessPool
from functools import partial
from queue import Queue
from typing import Optional

from pyqtgraph import QtCore, QtWidgets

from graphysio import dialogs, readdata, ui, utils
from graphysio.plotwidgets import TimeAxisItem, TSWidget


class MainUi(ui.Ui_MainWindow, QtWidgets.QMainWindow):
    setcoords = QtCore.pyqtSignal(float, float)

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self.setupUi(self)
        self.dircache = os.path.expanduser("~")

        self.dataq = Queue()
        self.pool = ProcessPool()
        self.datahandler = self.createNewPlotWithData

        self.tabWidget.tabCloseRequested.connect(self.closeTab)
        self.tabWidget.currentChanged.connect(self.tabChanged)

        launchNewPlot = partial(self.launchOpenFile, self.createNewPlotWithData)
        launchAppendPlot = partial(self.launchOpenFile, self.appendToPlotWithData)
        self.menuFile.addAction("New Plot", launchNewPlot)
        self.menuFile.addAction("Append to Plot", launchAppendPlot)

        if readdata.DwcReader.is_available:
            launchNewDwcPlot = partial(self.launchOpenDwc, self.createNewPlotWithData)
            self.menuFile.addAction("New Plot from DWC", launchNewDwcPlot)

        self.menuFile.addSeparator()
        self.menuFile.addAction("&Load plugin", self.errguard(utils.loadmodule))
        self.menuFile.addSeparator()
        self.menuFile.addAction("&Quit", self.close, QtCore.Qt.CTRL + QtCore.Qt.Key_Q)

        self.setcoords.connect(self.setCoords)

        # Launch timer to handle new plot data
        self.data_timer = QtCore.QTimer()
        self.data_timer.timeout.connect(self.read_plot_data)
        self.data_timer.setInterval(1000)  # Milliseconds
        self.data_timer.start()

    def closeEvent(self, event) -> None:
        self.pool.stop()
        self.pool.join()
        super().closeEvent(event)

    def setCoords(self, x, y) -> None:
        if TimeAxisItem.is_relative_time(x):
            timestr = TimeAxisItem.conv_relative(x)
        else:
            timestr = TimeAxisItem.conv_absolute(x, mainwindow=True)
        self.lblCoords.setText(f"Time: {timestr}, Value: {y}")

    def read_plot_data(self) -> None:
        while not self.dataq.empty():
            data = self.dataq.get()
            # Plotdata can be an object or a list of objects
            try:
                iter(data)
            except TypeError:
                data = [data]
            for plotdata in data:
                self.datahandler(plotdata)

    def print_exception(self, e) -> None:
        traceback.print_exc(file=sys.stdout)
        utils.displayError(e)

    def errguard(self, f):
        # Lift exceptions to UI reported errors
        def wrapped() -> None:
            try:
                f()
            except Exception as e:
                self.print_exception(e)

        return wrapped

    def addTab(self, widget: QtWidgets.QWidget, tabname: str) -> None:
        tabnames = []
        for i in range(self.tabWidget.count()):
            tabnames.append(self.tabWidget.tabText(i))
        if tabname in tabnames:
            tmptabname = tabname
            # Duplicate tab name. Add a number to the end
            for i in itertools.count():
                tmptabname = f"{tabname}_{i+1}"
                if tmptabname not in tabnames:
                    break
            tabname = tmptabname
        tabindex = self.tabWidget.addTab(widget, tabname)
        self.tabWidget.setCurrentIndex(tabindex)

    def closeTab(self, i) -> None:
        w = self.tabWidget.widget(i)
        self.tabWidget.removeTab(i)
        w.close()
        w.deleteLater()
        del w

    def tabChanged(self, tabid) -> None:
        destwidget = self.tabWidget.widget(tabid)
        if destwidget is None:
            return
        self.menuBar.clear()
        self.menuBar.addMenu(self.menuFile)
        for title, submenu in destwidget.menu.items():
            menu = self.menuBar.addMenu(title)
            for title, item in submenu.items():
                menu.addAction(title, item)

    def launchOpenDwc(self, datahandler) -> None:
        reader = readdata.DwcReader()
        reader.askUserInput()
        self.lblStatus.setText("Loading DWC...")
        future = self.pool.schedule(reader.get_plotdata)

        def cb(future) -> None:
            self.lblStatus.setText("Loading... done")
            plotdata = future.result()
            if plotdata:
                self.dataq.put(plotdata)

        self.datahandler = datahandler
        future.add_done_callback(cb)

    def launchOpenFile(self, datahandler) -> None:
        reader = readdata.FileReader()
        self.dircache = reader.askFile(self.dircache)
        self.lblStatus.setText("Loading File...")
        future = self.pool.schedule(reader.get_plotdata)

        def cb(future) -> None:
            self.lblStatus.setText("Loading... done")
            plotdata = future.result()
            if plotdata:
                self.dataq.put(plotdata)

        self.datahandler = datahandler
        future.add_done_callback(cb)

    def createNewPlotWithData(self, plotdata) -> None:
        properties = {"dircache": self.dircache}
        plotwidget = TSWidget(plotdata, parent=self, properties=properties)
        self.addTab(plotwidget, plotdata.name)

    def appendToPlotWithData(
        self,
        plotdata,
        destidx=None,
        do_timeshift: Optional[bool] = None,
    ) -> None:
        if destidx is None:
            plotwidget = self.tabWidget.currentWidget()
        else:
            plotwidget = self.tabWidget.widget(destidx)
        if plotwidget is None:
            utils.displayError("No plot selected.")
            return
        if do_timeshift is None:
            do_timeshift = dialogs.userConfirm(
                "Timeshift new curves to make the beginnings coincide?",
                title="Append to plot",
            )

        for fieldname in plotdata.data:
            newname = plotwidget.validateNewCurveName(fieldname)
            if newname is None:
                continue
            if newname != fieldname:
                plotdata.data = plotdata.data.rename(columns={fieldname: newname})

        plotwidget.appendData(plotdata, do_timeshift)
        plotwidget.properties["dircache"] = self.dircache



from copy import copy
from datetime import datetime
from functools import reduce
from math import floor
from typing import Dict

import numexpr as ne
import numpy as np
import pandas as pd
from pint import UnitRegistry
from scipy import interpolate, signal

from graphysio.structures import Filter, Parameter
from graphysio.utils import truncatevecs


class TF:
    def __init__(self, num, den, name="") -> None:
        self.num = num
        self.den = den
        self.name = name

    def discretize(self, samplerate):
        systf = (self.num, self.den)
        (dnum, dden, _) = signal.cont2discrete(systf, 1 / samplerate)
        return (np.squeeze(dnum), np.squeeze(dden))


interpkind = ["linear", "zero", "pchip", "nearest"]
Filters = {
    "Lowpass filter": Filter(
        name="lowpass",
        parameters=[
            Parameter("Cutoff frequency (Hz)", int),
            Parameter("Filter order", int),
        ],
    ),
    "Filter ventilation": Filter(name="ventilation", parameters=[]),
    "Savitzky-Golay": Filter(
        name="savgol",
        parameters=[
            Parameter("Window duration", "time"),
            Parameter("Polynomial order", int),
        ],
    ),
    "Interpolate": Filter(
        name="interp",
        parameters=[
            Parameter("New sampling rate (Hz)", float),
            Parameter("Interpolation type", interpkind),
        ],
    ),
    "Doppler cut": Filter(
        name="dopplercut",
        parameters=[Parameter("Minimum value", int)],
    ),
    "Differentiate": Filter(
        name="diff",
        parameters=[Parameter("Order", int), Parameter("Denominator time unit", str)],
    ),
    "Integrate": Filter(
        name="integrate",
        parameters=[Parameter("Window duration", "time")],
    ),
    "Lag": Filter(name="lag", parameters=[Parameter("Time delta", "time")]),
    "Normalize": Filter(name="norm1", parameters=[]),
    "Set start date/time": Filter(
        name="setdatetime",
        parameters=[Parameter("DateTime", datetime)],
    ),
    "Tolerant Normalize": Filter(name="norm2", parameters=[]),
    "Enter expression (variable = x)": Filter(
        name="expression",
        parameters=[Parameter("Expression", str)],
    ),
    "Strided Moving average": Filter(
        name="sma",
        parameters=[Parameter("Window duration", "time")],
    ),
}

FeetFilters = {
    "Short cycles": Filter(
        name="shortcycles",
        parameters=[Parameter("Minimum duration", "time")],
    ),
    "Extra feet": Filter(name="extrafeet", parameters=[]),
}

TFs: Dict[str, TF] = {}


def updateTFs() -> None:
    tflist = list(TFs.keys())
    if not tflist:
        return
    Filters["Transfer function"] = Filter(
        name="tf",
        parameters=[Parameter("Transfer function", tflist)],
    )


# -- Curve Filters --
def norm1(series, samplerate, parameters):
    series -= np.mean(series)
    series /= np.max(series) - np.min(series)
    newname = f"{series.name}-norm"
    return (series.rename(newname), None)


def norm2(series, samplerate, parameters):
    series -= np.mean(series)
    series /= np.std(series)
    newname = f"{series.name}-norm"
    return (series.rename(newname), None)


def expression(series, samplerate, parameters):
    (express,) = parameters
    try:
        filtered = ne.evaluate(express, local_dict={"x": series.to_numpy()})
    except Exception:
        filtered = None
    newname = f"{series.name}-filtered"
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, None)


def setdatetime(series, samplerate, parameters):
    (timestamp,) = parameters
    newseries = series.copy()
    diff = timestamp - newseries.index[0]
    newseries.index += diff
    newseries = newseries.rename(f"{series.name}-{timestamp}")
    return (newseries, None)


def sma(series, samplerate, parameters):
    (window_s,) = parameters
    serarr = series.to_numpy()
    serlen = len(serarr)
    winsize = int(window_s * samplerate)
    nwindows = floor(serlen / winsize)
    valstarts = np.arange(0, serlen - winsize, step=winsize)
    resiter = (
        np.mean(serarr[beg:end]) for beg, end in zip(valstarts, valstarts + winsize)
    )
    result = np.fromiter(resiter, dtype=np.float64, count=nwindows)
    locidx = winsize * np.arange(0, nwindows) + winsize // 2
    newname = f"{series.name}-sma{window_s}s"
    newseries = pd.Series(result, index=series.index[locidx], name=newname)
    newsamplerate = samplerate / winsize
    return (newseries, newsamplerate)


def savgol(series, samplerate, parameters):
    window, order = parameters
    window = np.floor(window * samplerate)
    if not window % 2:
        # window is even, we need odd
        window += 1
    filtered = signal.savgol_filter(series.values, int(window), order)
    newname = f"{series.name}-filtered"
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, None)


def lag(series, samplerate, parameters):
    (timedelta,) = parameters
    nindex = series.index.values - timedelta * 1e9
    newname = f"{series.name}-{timedelta}s"
    newseries = pd.Series(series.values, index=nindex, name=newname)
    return (newseries, None)


def tf(series, samplerate, parameters):
    (filtname,) = parameters
    tf = TFs[filtname]
    b, a = tf.discretize(samplerate)
    filtered = signal.lfilter(b, a, series)
    newname = f"{series.name}-{tf.name}"
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, None)


def lowpass(series, samplerate, parameters):
    Fc, order = parameters
    Wn = Fc * 2 / samplerate
    b, a = signal.butter(order, Wn)
    filtered = signal.lfilter(b, a, series)
    newname = f"{series.name}-lp{Fc}"
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, None)


def ventilation(series, samplerate, parameters):
    order = 12
    # Filter between 10 and 20 per minute
    bornes_hz = np.array([10, 20]) / 60
    Wn = [Fc * 2 / samplerate for Fc in bornes_hz]
    b, a = signal.butter(order, Wn, btype="bandstop")
    filtered = signal.filtfilt(b, a, series.values)
    newname = f"{series.name}-novent"
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, None)


def interp(series, samplerate, parameters):
    newsamplerate, method = parameters
    oldidx = series.index
    if method == "pchip":
        f = interpolate.PchipInterpolator(oldidx, series.values, extrapolate=True)
    else:
        f = interpolate.interp1d(
            oldidx,
            series.values,
            kind=method,
            fill_value="extrapolate",
        )
    step = 1e9 / newsamplerate  # 1e9 to convert Hz to ns
    newidx = np.arange(oldidx[0], oldidx[-1], step, dtype=np.int64)
    resampled = f(newidx)
    newname = f"{series.name}-{newsamplerate}Hz"
    newseries = pd.Series(resampled, index=newidx, name=newname)
    return (newseries, newsamplerate)


def dopplercut(series, samplerate, parameters):
    (minvel,) = parameters
    (notlow,) = (series < minvel).to_numpy().nonzero()
    newname = f"{series.name}-{minvel}+"
    newseries = series.rename(newname)
    newseries.iloc[notlow] = 0
    return (newseries, None)


def integrate(series, samplerate, parameters):
    (duration,) = parameters
    window = int(np.floor(duration * samplerate))
    integrated = series.rolling(window, center=True).sum()
    newname = f"{series.name}-integrate{duration}s"
    newseries = integrated.rename(newname)
    return (newseries, None)


def diff(series, samplerate, parameters):
    (order, timeunit) = parameters
    timeunit = timeunit.strip()
    dy = np.diff(series.values, order)
    dt = np.diff(series.index.values, order)
    ureg = UnitRegistry()
    ptu = ureg.parse_expression(timeunit)
    conv = ptu.to(ureg.nanosecond).magnitude
    dt = dt.astype(float) / conv
    diffed = dy / dt
    newname = f"{series.name}-diff{order}(/{timeunit})"
    newseries = pd.Series(diffed, index=series.index[order:], name=newname)
    return (newseries, None)


filtfuncs = {
    "savgol": savgol,
    "lag": lag,
    "tf": tf,
    "sma": sma,
    "lowpass": lowpass,
    "ventilation": ventilation,
    "interp": interp,
    "dopplercut": dopplercut,
    "integrate": integrate,
    "diff": diff,
    "norm1": norm1,
    "norm2": norm2,
    "expression": expression,
    "setdatetime": setdatetime,
}


def filter(curve, filtname, paramgetter):
    samplerate = curve.samplerate
    series = curve.series
    filt = Filters[filtname]
    parameters = map(paramgetter, filt.parameters)
    return filtfuncs[filt.name](series, samplerate, parameters)


# -- Feet Filters --
def shortcycles(feetdict, samplerate, parameters):
    starts = feetdict["start"]
    stops = feetdict["stop"]
    if len(stops) < 1:
        # No stop information
        msg = "No stop feet"
        raise ValueError(msg)
    (minimum_duration,) = parameters
    minimum_cycle_len = minimum_duration * 1e9  # Transform to ns
    cycle_durations = (stop - start for start, stop in zip(starts, stops))
    boolidx = [d >= minimum_cycle_len for d in cycle_durations]
    starts, stops, boolidx = truncatevecs([starts, stops, boolidx])
    feetdict["start"] = starts[boolidx]
    feetdict["stop"] = stops[boolidx]
    return feetdict


def extrafeet(feetdict, samplerate, parameters):
    starts = feetdict["start"]
    interval_quant_lo, median_interval, interval_quant_hi = (
        starts.to_series().diff().quantile([0.1, 0.5, 0.9])
    )
    interval_iqr = interval_quant_hi - interval_quant_lo

    # Construct a list of acceptable locations
    masks = []
    pos = starts[0]
    while True:
        lo = pos - interval_iqr // 2
        hi = pos + interval_iqr // 2
        pos += median_interval
        if lo > starts[-1]:
            break
        mask = np.ma.masked_outside(starts.to_numpy(), lo, hi).mask
        if not isinstance(mask, bool):
            masks.append(mask)

    # Apply a mask of acceptable locations to input
    mask = reduce(np.logical_and, masks)
    valid_starts = starts[~mask]
    feetdict["start"] = valid_starts
    return feetdict


feetfiltfuncs = {
    "shortcycles": shortcycles,
    "extrafeet": extrafeet,
}


def filter_feet(curve, filtname, paramgetter):
    samplerate = curve.samplerate
    feetdict = copy(curve.feetitem.indices)
    filt = FeetFilters[filtname]
    parameters = map(paramgetter, filt.parameters)
    return feetfiltfuncs[filt.name](feetdict, samplerate, parameters)



import string
from functools import partial

import numexpr as ne
import pandas as pd
from pyqtgraph.Qt import QtCore, QtWidgets

from graphysio import dialogs, transformations
from graphysio.algorithms import filters
from graphysio.plotwidgets import (
    LoopWidget,
    PlotWidget,
    POISelectorWidget,
    SpectrogramWidget,
)
from graphysio.structures import CycleId, Parameter, PlotData
from graphysio.writedata import exporter


class TSWidget(PlotWidget):
    def __init__(self, plotdata, parent=None, properties=None) -> None:
        if properties is None:
            properties = {}
        super().__init__(plotdata.name, parent=parent, properties=properties)
        self.exporter = exporter.TsExporter(self, plotdata.name)
        self.appendData(plotdata)

    def filterCurve(self, oldcurve, filtername, asnew=False) -> None:
        newseries, newsamplerate = filters.filter(
            oldcurve,
            filtername,
            dialogs.askUserValue,
        )
        if asnew:
            newname = self.validateNewCurveName(newseries.name)
            if newname != newseries.name:
                newseries = newseries.rename(newname)
            newcurve = self.addSeriesAsCurve(series=newseries)
            if newsamplerate and newcurve:
                newcurve.set_samplerate(newsamplerate)
        else:
            newseries = newseries.rename(oldcurve.series.name)
            oldcurve.replace_data(newseries)
            if newsamplerate:
                oldcurve.set_samplerate(newsamplerate)

    def filterFeet(self, curve, filtername, asnew=False) -> None:
        new_feetdict = filters.filter_feet(curve, filtername, dialogs.askUserValue)
        curve.feetitem.indices = new_feetdict
        curve.feetitem.render()

    def keyPressEvent(self, event) -> None:
        if event.key() == QtCore.Qt.Key_Delete:
            for curve in self.curves.values():
                curve.feetitem.removeSelection()

    # Menu Curves
    def showCurveList(self) -> None:
        dlgCurveSelection = dialogs.DlgCurveSelection(
            parent=self.parent,
            visible=self.curves.values(),
            hidden=self.hiddencurves,
        )

        def cb(result) -> None:
            currently_visible = set(self.curves.values())
            currently_hidden = self.hiddencurves
            allcurves = currently_visible | currently_hidden
            newly_visible, curveproperties = result
            newly_hidden = allcurves - newly_visible

            for c in curveproperties:
                self.applyCurveProperties(c, curveproperties[c])

            for c in newly_visible:
                self.addCurve(c)

            for c in newly_hidden:
                self.removeCurve(c)

            self.hiddencurves = newly_hidden
            self.rebuildLegend()

        dlgCurveSelection.dlgdata.connect(cb)
        dlgCurveSelection.exec_()

    def launchCycleDetection(self) -> None:
        dlgCycles = dialogs.DlgCycleDetection(parent=self.parent)

        def cb(choices) -> None:
            for curvename, choice in choices.items():
                curve = self.curves[curvename]
                curve.addFeet(CycleId(choice))

        dlgCycles.dlgdata.connect(cb)
        dlgCycles.exec_()

    def launchFilter(self, filterfeet) -> None:
        dlgFilter = dialogs.DlgFilter(parent=self.parent, filterfeet=filterfeet)

        def cb(result) -> None:
            createnew, curvechoices = result
            filterfunc = self.filterFeet if filterfeet else self.filterCurve
            for curvename, choice in curvechoices.items():
                if choice == "None":
                    continue
                curve = self.curves[curvename]
                filterfunc(curve, choice, asnew=createnew)

        dlgFilter.dlgdata.connect(cb)
        dlgFilter.exec_()

    def setDateTime(self) -> None:
        if len(self.curves) < 1:
            return
        sortedcurves = sorted(
            self.curves.values(),
            key=lambda curve: curve.series.index[0],
        )
        fstcurve = sortedcurves[0]
        curtimestamp = fstcurve.series.index[0]
        dlg = dialogs.DlgSetDateTime(prevdatetime=curtimestamp)
        dlg.exec_()
        newtimestamp = dlg.dlgdata
        if not newtimestamp:
            return
        offset = newtimestamp - curtimestamp
        for curve in self.curves.values():
            curve.series.index += offset
            curve.render()

    def launchTransformation(self) -> None:
        param = Parameter(
            "Choose Transformation",
            list(transformations.Transformations.keys()),
        )
        qresult = dialogs.askUserValue(param)
        if qresult is None:
            return
        trans = transformations.Transformations[qresult]
        plotdatas = trans(self)
        for plotdata in plotdatas:
            self.parent.appendToPlotWithData(plotdata, do_timeshift=False)

    # Menu Plot
    def launchPOIWidget(self) -> None:
        curvenames = list(self.curves.keys())
        q = Parameter("Select Curve", curvenames)
        curvename = dialogs.askUserValue(q)
        if not curvename:
            return
        curve = self.curves[curvename]
        poiselector = POISelectorWidget(
            curve.series,
            parent=self.parent,
            properties=self.properties,
        )
        self.parent.addTab(poiselector, curve.name())

    def launchSpectrogram(self) -> None:
        curvenames = list(self.curves.keys())
        q = Parameter("Select Curve", curvenames)
        curvename = dialogs.askUserValue(q)
        q = Parameter("Enter time window", "time")
        window = dialogs.askUserValue(q)
        if not curvename:
            return
        curve = self.curves[curvename]
        spectro = SpectrogramWidget(
            curve.series,
            curve.samplerate,
            window,
            parent=self.parent,
        )
        self.parent.addTab(spectro, curve.name())

    def launchCurveAlgebra(self) -> None:
        curvecorr = dict(zip(string.ascii_lowercase, self.curves.keys()))
        dlgCurveAlgebra = dialogs.DlgCurveAlgebra(self, curvecorr)

        def cb(formula) -> None:
            symbols = list(ne.NumExpr(formula).input_names)
            curvenames = [curvecorr[x] for x in symbols]
            sers = [self.curves[c].series for c in curvenames]
            if not len(sers):
                return

            argsdf = pd.concat(sers, axis=1, keys=symbols)
            argsdf = argsdf.interpolate()
            args = argsdf.to_dict(orient="series")

            newvals = ne.evaluate(formula, local_dict=args)
            newname = self.validateNewCurveName(formula, True)
            newseries = pd.Series(
                newvals,
                index=argsdf.index,
                name=newname,
                dtype="float64",
            )
            self.addSeriesAsCurve(newseries)

        dlgCurveAlgebra.dlgdata.connect(cb)
        dlgCurveAlgebra.exec_()

    # Menu Selection
    def launchNewPlotFromSelection(self) -> None:
        xmin, xmax = self.vbrange
        sers = []
        for c in self.curves.values():
            series = c.series.loc[xmin:xmax]
            sers.append(series)
        df = pd.concat(sers, axis=1, keys=[s.name for s in sers])
        newname = f"{self.name}-sub"
        plotdata = PlotData(data=df, name=newname)
        self.parent.createNewPlotWithData(plotdata)

    def launchAppendToPlotFromSelection(self) -> None:
        tabWidget = self.parent.tabWidget
        ntabs = tabWidget.count()
        tabdict = {tabWidget.tabText(idx): idx for idx in range(ntabs)}
        desttabname, ok = QtWidgets.QInputDialog.getItem(
            self,
            "Select destination",
            "Destination plot",
            list(tabdict.keys()),
            editable=False,
        )
        if not ok:
            return
        xmin, xmax = self.vbrange
        sers = []
        for c in self.curves.values():
            series = c.series.loc[xmin:xmax]
            sers.append(series)
        df = pd.concat(sers, axis=1, keys=[s.name for s in sers])
        newname = f"{self.name}-sub"
        plotdata = PlotData(data=df, name=newname)
        self.parent.appendToPlotWithData(plotdata, destidx=tabdict[desttabname])

    def launchLoop(self) -> None:
        dlgSetupPU = dialogs.DlgSetupPULoop(self, parent=self.parent)

        def cb(result) -> None:
            uname, pname = result
            u = self.curves[uname]
            p = self.curves[pname]
            loopwidget = LoopWidget(u, p, self.vbrange, parent=self.parent)
            newname = f"{self.name}-{p.name()}-loops"
            self.parent.addTab(loopwidget, newname)

        dlgSetupPU.dlgdata.connect(cb)
        dlgSetupPU.exec_()

    @property
    def menu(self):
        mcurves = {
            "Visible Curves": self.showCurveList,
            "Cycle Detection": self.launchCycleDetection,
            "Filter Curve": partial(self.launchFilter, filterfeet=False),
            "Filter Feet": partial(self.launchFilter, filterfeet=True),
            "Transformation": self.launchTransformation,
            "Set Date / Time": self.setDateTime,
        }
        mselect = {
            "As new plot": self.launchNewPlotFromSelection,
            "Append to other plot": self.launchAppendToPlotFromSelection,
            "Generate PU-Loops": self.launchLoop,
        }
        mplot = {
            "POI Selector": self.launchPOIWidget,
            "Spectrogram": self.launchSpectrogram,
            "Curve Algebra": self.launchCurveAlgebra,
        }
        mexport = {
            "Curves": self.exporter.curves,
            "Time info": self.exporter.periods,
            "Cycle info": self.exporter.cyclepoints,
            "Cycles to directory": self.exporter.cycles,
        }
        return {
            "Curves": mcurves,
            "Plot": mplot,
            "Seletion": mselect,
            "Export": mexport,
        }



# Form implementation generated from reading ui file 'dwcopen.ui'
#
# Created by: PyQt6 UI code generator 6.7.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from pyqtgraph import QtCore, QtGui, QtWidgets


class Ui_DWCOpen(object):
    def setupUi(self, DWCOpen):
        DWCOpen.setObjectName("DWCOpen")
        DWCOpen.resize(564, 511)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(DWCOpen)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.txtPatientId = QtWidgets.QLineEdit(parent=DWCOpen)
        self.txtPatientId.setObjectName("txtPatientId")
        self.gridLayout.addWidget(self.txtPatientId, 0, 1, 1, 1)
        self.label = QtWidgets.QLabel(parent=DWCOpen)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.searchButton = QtWidgets.QPushButton(parent=DWCOpen)
        self.searchButton.setObjectName("searchButton")
        self.gridLayout.addWidget(self.searchButton, 0, 2, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.lblFound = QtWidgets.QLabel(parent=DWCOpen)
        self.lblFound.setText("")
        self.lblFound.setObjectName("lblFound")
        self.verticalLayout_2.addWidget(self.lblFound)
        self.cmbTypeofData = QtWidgets.QComboBox(parent=DWCOpen)
        self.cmbTypeofData.setObjectName("cmbTypeofData")
        self.cmbTypeofData.addItem("")
        self.cmbTypeofData.addItem("")
        self.verticalLayout_2.addWidget(self.cmbTypeofData)
        self.lstLabels = QtWidgets.QListWidget(parent=DWCOpen)
        self.lstLabels.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.MultiSelection)
        self.lstLabels.setObjectName("lstLabels")
        self.verticalLayout_2.addWidget(self.lstLabels)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_2 = QtWidgets.QLabel(parent=DWCOpen)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_2.addWidget(self.label_2)
        self.dtFrom = QtWidgets.QDateTimeEdit(parent=DWCOpen)
        self.dtFrom.setCalendarPopup(True)
        self.dtFrom.setObjectName("dtFrom")
        self.horizontalLayout_2.addWidget(self.dtFrom)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_3 = QtWidgets.QLabel(parent=DWCOpen)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_3.addWidget(self.label_3)
        self.dtTo = QtWidgets.QDateTimeEdit(parent=DWCOpen)
        self.dtTo.setCalendarPopup(True)
        self.dtTo.setObjectName("dtTo")
        self.horizontalLayout_3.addWidget(self.dtTo)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.okButton = QtWidgets.QPushButton(parent=DWCOpen)
        self.okButton.setObjectName("okButton")
        self.horizontalLayout.addWidget(self.okButton)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.cancelButton = QtWidgets.QPushButton(parent=DWCOpen)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.verticalLayout_2.addLayout(self.horizontalLayout)

        self.retranslateUi(DWCOpen)
        self.cmbTypeofData.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(DWCOpen)

    def retranslateUi(self, DWCOpen):
        _translate = QtCore.QCoreApplication.translate
        DWCOpen.setWindowTitle(_translate("DWCOpen", "Load DWC data"))
        self.label.setText(_translate("DWCOpen", "Patient ID"))
        self.searchButton.setText(_translate("DWCOpen", "Search"))
        self.cmbTypeofData.setItemText(0, _translate("DWCOpen", "Numerics"))
        self.cmbTypeofData.setItemText(1, _translate("DWCOpen", "Waves"))
        self.label_2.setText(_translate("DWCOpen", "From"))
        self.label_3.setText(_translate("DWCOpen", "To"))
        self.okButton.setText(_translate("DWCOpen", "Ok"))
        self.cancelButton.setText(_translate("DWCOpen", "Cancel"))



from datetime import datetime
from typing import List

import numpy as np
from scipy import interpolate

from graphysio.dialogs import askUserValue
from graphysio.plotwidgets.curves import CurveItem
from graphysio.structures import Parameter

try:
    import pyedflib
except ImportError:
    is_available = False
else:
    is_available = True


def interp_series(s, samplerate, start, stop):
    f = interpolate.interp1d(
        s.index,
        s.values,
        assume_sorted=True,
        copy=False,
        fill_value="extrapolate",
    )
    # Account for periods of missing data.
    duration_s = (stop - start) * 1e-9
    npoints = int(duration_s * samplerate)
    newidx = np.linspace(start, stop, num=npoints)
    return f(newidx)


def curves_to_edf(
    curves: List[CurveItem],
    filepath: str,
    index_label: str = "timens",
) -> None:
    headers = []
    signals = []

    beginns = min(c.series.index[0] for c in curves)
    endns = max(c.series.index[-1] for c in curves)
    begindt = datetime.fromtimestamp(beginns * 1e-9)

    # Use global min / max of values since some viewers need this (edfbrowser)
    physmax = max(c.series.max() for c in curves)
    physmin = min(c.series.min() for c in curves)

    # Ask the user for the physical dimension shared by all curves
    dim = askUserValue(Parameter("Enter physical dimension", str))

    for c in curves:
        s = c.series
        header = {
            "label": c.name(),
            "sample_rate": c.samplerate,
            "physical_max": physmax,
            "physical_min": physmin,
            "digital_max": 32767,
            "digital_min": -32768,
            "transducer": "",
            "prefilter": "",
            "dimension": dim,
        }
        headers.append(header)
        resampled = interp_series(s, c.samplerate, beginns, endns)
        signals.append(resampled)

    edf = pyedflib.EdfWriter(
        str(filepath),
        len(signals),
        file_type=pyedflib.FILETYPE_EDFPLUS,
    )
    edf.setStartdatetime(begindt)
    edf.setSignalHeaders(headers)
    edf.writeSamples(signals)
    edf.close()


curves_to_edf.is_available = is_available



from functools import partial
from typing import Optional

import numpy as np
import pandas as pd
import pyqtgraph as pg
from physiocurve.pandas import ECG, Pressure
from pyqtgraph.Qt import QtCore, QtGui

from graphysio import utils
from graphysio.algorithms import waveform
from graphysio.structures import CycleId
from graphysio.utils import estimateSampleRate


class CurveItem(pg.PlotDataItem):
    visible = QtCore.pyqtSignal()
    invisible = QtCore.pyqtSignal()

    def __init__(self, series, parent, pen=None) -> None:
        self.parent = parent
        self.series = self.sanitize_data(series)
        if self.series is None:
            msg = "Not enough data"
            raise ValueError(msg)
        self.samplerate = estimateSampleRate(self.series)
        if pen is None:
            pen = QtGui.QColor(QtCore.Qt.black)
        super().__init__(name=series.name, pen=pen, antialias=True)
        self.render()

    def replace_data(self, newdata) -> None:
        newdata = self.sanitize_data(newdata)
        if newdata is None:
            # XXX report error
            return
        self.clear()
        self.series = newdata.rename(self.opts["name"])
        self.samplerate = estimateSampleRate(self.series)
        self.render()

    def extend(self, newseries) -> None:
        merged = pd.concat([self.series, newseries])
        newdata = self.sanitize_data(merged)
        if newdata is None:
            # XXX report error
            return
        self.series = newdata
        self.render()

    def render(self) -> None:
        self.setData(x=self.series.index.to_numpy(), y=self.series.to_numpy())

    def set_samplerate(self, newsamplerate) -> None:
        self.samplerate = newsamplerate

    def rename(self, newname: str) -> None:
        self.opts["name"] = newname
        self.series = self.series.rename(newname)

    def sanitize_data(self, series: pd.Series) -> Optional[pd.Series]:
        series = series.dropna()
        series = series.sort_index()
        # Less than 2 points are not visible
        if len(series) < 2:
            return None
        # Make timestamp unique and use mean of values on duplicates
        return series.groupby(level=0).mean()


class POIItem(pg.ScatterPlotItem):
    sym = {
        "start": "star",
        "stop": "s",
        "diastole": "t1",
        "systole": "t",
        "dicrotic": "d",
        "qwave": "t1",
        "rwave": "x",
        "swave": "t",
        "twave": "o",
    }
    # Symbols = OrderedDict([(name, QtGui.QPainterPath()) for name in
    # ['o', 's', 't', 't1', 't2', 't3','d', '+', 'x', 'p', 'h', 'star']])

    def __init__(self, parent, name, pen=None) -> None:
        super().__init__(pen=pen, name=name)
        self.parent = parent
        self.indices = {}
        self.selected = []
        self.resym = {value: key for key, value in self.sym.items()}
        self.render()

    def addPointsByLocation(self, key, locations) -> None:
        if key not in self.indices:
            self.indices[key] = pd.Index([])
        oldidx = self.indices[key]
        newidx = oldidx.append(pd.Index(locations))
        self.indices[key] = newidx.unique().sort_values()
        self.render()

    def removePointsByLocation(self, key, locations) -> None:
        if key not in self.indices:
            return
        oldidx = self.indices[key]
        dellocs = oldidx.get_indexer(locations, method="nearest")
        newidx = oldidx.delete(dellocs)
        self.indices[key] = newidx
        self.render()

    def removePoints(self, points) -> None:
        for point in points:
            try:
                sym = self.resym[point.symbol()]
                idx = self.indices[sym]
            except KeyError as e:
                msg = "Point not found"
                raise KeyError(msg) from e
            nidx = idx.get_indexer([point.pos().x()], method="nearest")
            self.indices[sym] = idx.delete(nidx)
        self.render()

    def render(self) -> None:
        data = []
        for key, idx in self.indices.items():
            if len(idx) < 1:
                continue
            idxnona = idx.dropna()
            points = self.parent.series.loc[idxnona]
            tmp = pd.DataFrame({"points": points, "sym": self.sym[key]}, index=idxnona)
            data.append(tmp)
        if len(data) < 1:
            self.clear()
            return
        feet = pd.concat(data).sort_index()
        self.setData(
            x=feet.index.values,
            y=feet["points"].values,
            symbol=feet["sym"].values,
        )

    def isPointSelected(self, point):
        return point in self.selected

    def selectPoint(self, point) -> None:
        if not self.isPointSelected(point):
            self.selected.append(point)
        point.setPen("r")
        point.setBrush("r")

    def unselectPoint(self, point) -> None:
        if self.isPointSelected(point):
            self.selected.remove(point)
        point.resetPen()
        point.resetBrush()

    def removeSelection(self) -> None:
        self.removePoints(self.selected)
        self.selected = []

    def rename(self, newname: str) -> None:
        self.opts["name"] = newname


class CurveItemWithPOI(CurveItem):
    visible = QtCore.pyqtSignal()
    invisible = QtCore.pyqtSignal()

    @staticmethod
    def sigPointClicked(feetitem, points) -> None:
        point = points[0]  # only one point per click
        if not feetitem.isPointSelected(point):
            feetitem.selectPoint(point)
        else:
            feetitem.unselectPoint(point)

    def __init__(self, series, parent, pen=None) -> None:
        super().__init__(series, parent, pen)
        feetname = f"{series.name}-feet"
        self.feetitem = POIItem(self, name=feetname, pen=pen)
        parent.addItem(self.feetitem)
        self.feetitem.sigClicked.connect(self.sigPointClicked)
        self.visible.connect(self.__becameVisible)
        self.invisible.connect(self.__becameInvisible)

    def __becameVisible(self) -> None:
        if self.feetitem not in self.parent.listDataItems():
            self.parent.addItem(self.feetitem)
        self.render()
        self.feetitem.render()

    def __becameInvisible(self) -> None:
        self.parent.removeItem(self.feetitem)

    def addFeet(self, cycleid) -> None:
        if cycleid is CycleId.none:
            return
        elif cycleid is CycleId.velocity:
            starts, stops = waveform.findFlowCycles(self)
            self.feetitem.indices["start"] = starts
            self.feetitem.indices["stop"] = stops
        elif cycleid is CycleId.foot:
            foot = waveform.findPressureFeet(self)
            self.feetitem.indices["start"] = foot
        elif cycleid is CycleId.pressure:
            if "start" not in self.feetitem.indices:
                self.addFeet(CycleId.foot)
            dia, sbp, dic = waveform.findPressureFull(self)
            self.feetitem.indices["diastole"] = dia
            self.feetitem.indices["systole"] = sbp
            self.feetitem.indices["dicrotic"] = dic
        elif cycleid is CycleId.rwave:
            ecg = ECG(self.series)
            self.feetitem.indices["rwave"] = ecg.idxrwave
        elif cycleid is CycleId.ecg:
            ecg = ECG(self.series)
            self.feetitem.indices["start"] = ecg.idxpwave
            self.feetitem.indices["qwave"] = ecg.idxqwave
            self.feetitem.indices["rwave"] = ecg.idxrwave
            self.feetitem.indices["swave"] = ecg.idxswave
            self.feetitem.indices["twave"] = ecg.idxtwave
        elif cycleid is CycleId.foottan:
            p = Pressure(self.series)
            self.feetitem.indices["start"] = p.idxtanfeet
        elif cycleid is CycleId.pressurebis:
            p = Pressure(self.series)
            self.feetitem.indices["start"] = p.idxfeet
            self.feetitem.indices["diastole"] = p.idxdia
            self.feetitem.indices["systole"] = p.idxsys
            self.feetitem.indices["dicrotic"] = p.idxdic
        else:
            raise ValueError(cycleid)
        self.feetitem.render()

    def getCycleIndices(self, vrange=None):
        s = self.series
        clip = partial(utils.clip, vrange=vrange)
        hasstarts = ("start" in self.feetitem.indices) and self.feetitem.indices[
            "start"
        ].size > 0
        hasstops = ("stop" in self.feetitem.indices) and self.feetitem.indices[
            "stop"
        ].size > 0
        if vrange:
            xmin, xmax = vrange
        else:
            xmin = s.index[0]
            xmax = s.index[-1]
        if not hasstarts:
            # We have no feet, treat the whole signal as one cycle
            locs = s.index.get_indexer([xmin, xmax], method="nearest")
            indices = s.index[locs]
            begins, ends = (np.array([i]) for i in indices)
        elif not hasstops:
            # We have no stops, starts serve as stops for previous cycle
            begins = clip(self.feetitem.indices["start"].values)
            endloc = s.index.get_indexer([xmax], method="nearest")
            end = s.index[endloc]
            ends = np.append(begins[1:], end)
        else:
            # We have starts and stops, use them
            begins = self.feetitem.indices["start"].values
            ends = self.feetitem.indices["stop"].values
            begins, ends = map(clip, [begins, ends])

        # Handle the case where we start in the middle of a cycle
        while ends[0] <= begins[0]:
            ends = ends[1:]

        begins, ends = utils.truncatevecs([begins, ends])
        durations = ends - begins
        return (begins, durations)

    def getFeetPoints(self, feetname) -> Optional[pd.Series]:
        if feetname not in self.feetitem.indices:
            return None
        feetidx = self.feetitem.indices[feetname]
        feetnona = feetidx[pd.notnull(feetidx)]
        return self.series.loc[feetnona]



from graphysio.structures import PlotData


class BaseReader:
    def __init__(self) -> None:
        self.userdata = {}

    def set_data(self, data) -> None:
        self.userdata.update(data)

    def askUserInput(self) -> None:
        pass

    def __call__(self) -> PlotData:
        raise NotImplementedError



import pandas as pd
from pandas.api.types import is_datetime64_any_dtype

from graphysio.dialogs import DlgListChoice
from graphysio.readdata.baseclass import BaseReader
from graphysio.structures import PlotData

try:
    import pyarrow.parquet as pa
except ImportError:
    is_available = False
else:
    is_available = True


class ParquetReader(BaseReader):
    is_available = is_available

    def askUserInput(self) -> None:
        filepath = self.userdata["filepath"]
        s = pa.read_schema(filepath)
        colnames = s.names

        def cb(columns) -> None:
            self.userdata["columns"] = columns

        dlgchoice = DlgListChoice(colnames, "Open Parquet", "Choose curves to load")
        dlgchoice.dlgdata.connect(cb)
        dlgchoice.exec_()

    def __call__(self) -> PlotData:
        filepath = self.userdata["filepath"]
        data = pd.read_parquet(filepath, columns=self.userdata["columns"])

        data = data.dropna(axis="columns", how="all")
        data = data.sort_index()

        if is_datetime64_any_dtype(data.index):
            data.index = data.index.tz_localize(None)
        data.index = data.index.astype("datetime64[ns]").astype("int")

        return PlotData(data=data, filepath=filepath)



from graphysio.dialogs import DlgDWCOpen
from graphysio.readdata.baseclass import BaseReader
from graphysio.structures import PlotData
from pandas.api.types import is_datetime64_any_dtype
from typing import List

try:
    import dwclib
except ImportError:
    is_available = False
else:
    is_available = True


class DwcReader(BaseReader):
    is_available = is_available

    def askUserInput(self) -> None:
        def cb(reqdata) -> None:
            self.userdata = reqdata

        dlgchoice = DlgDWCOpen(dwclib.read_patient)
        dlgchoice.dlgdata.connect(cb)
        dlgchoice.exec_()

    def get_plotdata(self) -> PlotData | List[PlotData]:
        if not self.userdata:
            return []
        if self.userdata["type"] == "numerics":
            df = dwclib.read_numerics(
                patientids=self.userdata["patientid"],
                dtbegin=self.userdata["from"],
                dtend=self.userdata["to"],
                sublabels=self.userdata["items"],
            )
        elif self.userdata["type"] == "waves":
            df = dwclib.read_waves(
                patientid=self.userdata["patientid"],
                dtbegin=self.userdata["from"],
                dtend=self.userdata["to"],
                labels=self.userdata["items"],
            )
        else:
            raise ValueError("wrong data request type")

        if is_datetime64_any_dtype(df.index):
            df.index = df.index.tz_localize(None)
        df.index = df.index.astype("datetime64[ns]").astype("int")

        return PlotData(data=df, name=str(self.userdata["patientid"]))



from enum import Enum
from functools import partial

import pandas as pd
import pyqtgraph as pg
from pyqtgraph import QtWidgets

from graphysio import ui
from graphysio.algorithms.filters import savgol
from graphysio.algorithms.waveform import findPOIGreedy
from graphysio.plotwidgets import PlotWidget
from graphysio.structures import PlotData
from graphysio.writedata import exporter


class FixIndex(Enum):
    disabled = "Disabled"
    minimum = "Local Minimum"
    maximum = "Local Maximum"
    sndderiv = "2nd derivative peak"


class POISelectorWidget(ui.Ui_POISelectorWidget, QtWidgets.QWidget):
    @staticmethod
    def buttonClicked(self, qbutton) -> None:
        fixval = FixIndex(qbutton.text())
        self.poiselectorwidget.fixvalue = fixval

    def __init__(self, series, parent, properties=None) -> None:
        super().__init__(parent=parent)
        self.setupUi(self)
        self.parent = parent
        if properties is None:
            properties = {}

        self.poiselectorplot = POISelectorPlot(
            series,
            parent=self,
            properties=properties,
        )
        self.horizontalLayout.addWidget(self.poiselectorplot)

        buttonClicked = partial(self.buttonClicked, self)
        self.buttonGroup.buttonClicked.connect(buttonClicked)

    def loadPOI(self, plotdata) -> None:
        columnname, ok = QtWidgets.QInputDialog.getItem(
            self,
            "Select POI series",
            "Load POI",
            plotdata.data.columns,
            editable=False,
        )
        if not ok:
            return
        poiseries = plotdata.data[columnname].dropna().index
        self.poiselectorplot.curve.feetitem.addPointsByLocation(
            self.poiselectorplot.pointkey,
            poiseries,
        )

    def launchNewPlotFromPOIs(self) -> None:
        srcseries = self.poiselectorplot.curve.series
        poiidx = self.poiselectorplot.curve.feetitem.indices[
            self.poiselectorplot.pointkey
        ]
        pois = srcseries[poiidx.dropna()].rename("poi")
        sname = f"{srcseries.name}-poi"
        df = pd.DataFrame({sname: pois})
        plotdata = PlotData(data=df, name=sname)
        self.parent.createNewPlotWithData(plotdata)

    @property
    def menu(self):
        return {
            "Plot": {
                "Import POIs": partial(
                    self.parent.launchOpenFile,
                    datahandler=self.loadPOI,
                ),
                "POIs to New Plot": self.launchNewPlotFromPOIs,
            },
            "Export": {"POI to CSV": self.poiselectorplot.exporter.poi},
        }


class POISelectorPlot(PlotWidget):
    pointkey = "point"

    @staticmethod
    def mouseMoved(self, evt) -> None:
        pos = evt[0]  # using signal proxy turns original arguments into a tuple
        if self.sceneBoundingRect().contains(pos):
            mousePoint = self.getViewBox().mapSceneToView(pos)
            self.vLine.setPos(mousePoint.x())

    @staticmethod
    def clicked(self, evt) -> None:
        button = evt.button()
        pos = self.vLine.value()
        if button == 1:
            correctedpos = self.fixpos(pos)
            self.curve.feetitem.addPointsByLocation(self.pointkey, [correctedpos])
        elif button == 2:
            self.curve.feetitem.removePointsByLocation(self.pointkey, [pos])

    def __init__(self, series, parent, properties=None) -> None:
        super().__init__(name=series.name, parent=parent)
        self.vb.setMouseMode(self.vb.PanMode)
        self.setMenuEnabled(False)
        self.properties = properties if properties is not None else {}

        self.__sndderiv = None
        self.fixvalue = FixIndex.disabled

        self.curve = self.addSeriesAsCurve(series)
        if self.curve is None:
            msg = "Not enough data"
            raise ValueError(msg)
        self.exporter = exporter.POIExporter(self, self.name)

        pen = pg.mkPen("k", width=2)
        self.vLine = pg.InfiniteLine(angle=90, movable=False, pen=pen)
        self.addItem(self.vLine, ignoreBounds=True)

        clicked = partial(self.clicked, self)
        self.scene().sigMouseClicked.connect(clicked)

    def fixpos(self, pos):
        # Need to go through get_loc, otherwise strange things happen
        xmin, xmax = self.vbrange
        if self.fixvalue is FixIndex.minimum:
            s = self.curve.series.loc[xmin:xmax]
            posprop = findPOIGreedy(s, pos, "min")
            fixedposloc = s.index.get_loc(posprop, method="nearest")
        elif self.fixvalue is FixIndex.maximum:
            s = self.curve.series.loc[xmin:xmax]
            posprop = findPOIGreedy(s, pos, "max")
            fixedposloc = s.index.get_loc(posprop, method="nearest")
        elif self.fixvalue is FixIndex.sndderiv:
            s = self.sndderiv.loc[xmin:xmax]
            posprop = findPOIGreedy(s, pos, "max")
            fixedposloc = s.index.get_loc(posprop, method="nearest")
        else:
            s = self.curve.series.loc[xmin:xmax]
            fixedposloc = s.index.get_loc(pos, method="nearest")
        return s.index[fixedposloc]

    @property
    def sndderiv(self):
        if self.__sndderiv is None:
            sndderiv = self.curve.series.diff().diff().iloc[1:]
            sndderiv, _ = savgol(sndderiv, self.curve.samplerate, (0.16, 2))
            self.__sndderiv = sndderiv.dropna()
        return self.__sndderiv



import sys

from pyqtgraph.Qt import QtWidgets

from graphysio.mainui import MainUi


def main() -> None:
    app = QtWidgets.QApplication(sys.argv)

    winmain = MainUi()
    winmain.show()

    sys.exit(app.exec_())



import argparse

from pyshortcuts import make_shortcut

from graphysio.main import main as graphysio_main


def mk_shortscuts() -> None:
    pycmd = "_ -m graphysio"
    make_shortcut(pycmd, name="GraPhysio", terminal=False)


if __name__ == "__main__":
    parser = argparse.ArgumentParser("GraPhysio")
    parser.add_argument(
        "-S",
        "--shortcuts",
        action="store_true",
        help="Install Desktop shortcuts.",
    )
    args = parser.parse_args()

    if args.shortcuts:
        mk_shortscuts()
    else:
        graphysio_main()



"""Graphical time series visualizer and analyzer."""

__all__ = [
    "algorithms",
    "dialogs",
    "exporter",
    "legend",
    "mainui",
    "puplot",
    "tsplot",
    "utils",
    "types",
    "ui",
    "transformations",
]



from typing import List

from graphysio.dialogs import askUserValue
from graphysio.plotwidgets import PlotWidget
from graphysio.plotwidgets.curves import CurveItemWithPOI
from graphysio.structures import Parameter, PlotData


def get_feet_to_curve(plotwidget: PlotWidget) -> List[CurveItemWithPOI]:
    feetitemhash = {}
    for curve in plotwidget.curves.values():
        feetitemhash.update(
            {
                f"{curve.name()}-{feetname}": (curve, feetname)
                for feetname in curve.feetitem.indices
            },
        )
    param = Parameter("Choose points to create curve", list(feetitemhash.keys()))
    qresult = askUserValue(param)
    curve, feetname = feetitemhash[qresult]

    out_series = curve.getFeetPoints(feetname)
    sname = qresult

    plotdata = PlotData(data=out_series, name=sname)
    return [plotdata]



import csv
import os
from itertools import zip_longest

import pandas as pd

from graphysio import writedata
from graphysio.dialogs import DlgPeriodExport, askDirPath, askSaveFilePath
from graphysio.utils import sanitize_filename

file_filters = ";;".join(
    [f"{ext.upper()} files (*.{ext})" for ext in writedata.curve_writers],
)


class TsExporter:
    periodfields = ["patient", "begin", "end", "periodid", "comment"]

    def __init__(self, parent, name) -> None:
        self.parent = parent
        self.name = name
        try:
            self.outdir = parent.properties["dircache"]
        except KeyError:
            self.outdir = os.path.expanduser("~")

    def curves(self) -> None:
        filepath, ext = askSaveFilePath(
            "Export to",
            f"{self.name}.csv",
            self.outdir,
            filter=file_filters,
        )
        if filepath is None:
            return
        self.outdir = os.path.dirname(filepath)
        curves = list(self.parent.curves.values())
        export_func = writedata.curve_writers[ext]
        export_func(curves, filepath)

    def periods(self) -> None:
        xmin, xmax = self.parent.vbrange
        dlg = DlgPeriodExport(
            begin=xmin,
            end=xmax,
            patient=self.name,
            directory=self.outdir,
        )

        def cb(result) -> None:
            patient, comment, periodname, filepath = result
            self.name = patient
            self.outdir = os.path.dirname(filepath)

            with open(filepath, "a", newline="") as csvfile:
                writer = csv.DictWriter(
                    csvfile,
                    fieldnames=self.periodfields,
                    quoting=csv.QUOTE_MINIMAL,
                )
                if not os.path.exists(filepath):
                    writer.writeheader()
                writer.writerow(
                    {
                        "patient": patient,
                        "begin": xmin,
                        "end": xmax,
                        "periodid": periodname,
                        "comment": comment,
                    },
                )

        dlg.dlgdata.connect(cb)
        dlg.exec_()

    def cycles(self) -> None:
        outdir = askDirPath("Export to", self.outdir)
        if outdir is None:
            # Cancel pressed
            return

        # Some non trivial manipulations to get the cycles from all
        # curves, then reorganize to group by the n-th cycle from each
        # curve and put those cycles into a dataframe for export.
        def getCurveCycles(curve):
            cycleIdx = curve.getCycleIndices()
            return (curve.series.loc[b : b + d] for b, d in zip(*cycleIdx))

        curves = self.parent.curves.values()
        allByCurve = (getCurveCycles(curve) for curve in curves)
        allByCycle = zip_longest(*allByCurve)

        for n, cycle in enumerate(allByCycle):
            idxstart = None
            for s in cycle:
                # Make all series start at the same point
                if s is None or len(s) < 1:
                    continue
                if idxstart is None:
                    idxstart = s.index[0]
                idxdelta = s.index[0] - idxstart
                s.index -= idxdelta
            df = pd.concat(cycle, axis=1)
            df["datetime"] = pd.to_datetime(df.index, unit="ns")

            filename = sanitize_filename(f"{self.name}-{n+1}.csv")
            filepath = os.path.join(self.outdir, filename)
            df.to_csv(
                filepath,
                date_format="%Y-%m-%d %H:%M:%S.%f",
                index_label="timens",
            )

    def cyclepoints(self) -> None:
        filepath, _ = askSaveFilePath("Export to", f"{self.name}-feet.csv", self.outdir)
        if filepath is None:
            # Cancel pressed
            return
        self.outdir = os.path.dirname(filepath)

        feetseries = []
        for c in self.parent.curves.values():
            for k, v in c.feetitem.indices.items():
                feetseries.append(pd.Series(v, name=f"{c.name()}-{k}"))
        df = pd.concat(feetseries, axis=1).astype("Int64")
        df.to_csv(filepath, index_label="idx")


class PuExporter:
    def __init__(self, parent, name) -> None:
        self.parent = parent
        self.name = name
        try:
            self.outdir = parent.properties["dircache"]
        except KeyError:
            self.outdir = os.path.expanduser("~")

    def exportloops(self) -> None:
        outdir = askDirPath("Export to", self.outdir)
        if outdir is None:
            # Cancel pressed
            return
        self.outdir = outdir
        self.writetable()
        self.writeloops()

    def writetable(self) -> None:
        data = []
        for loop in self.parent.loops:
            alpha, beta, gala = loop.angles
            delay = loop.offset / 1e6
            tmpdict = {"alpha": alpha, "beta": beta, "gala": gala, "delay": delay}
            data.append(tmpdict)
        df = pd.DataFrame(data)
        df.index += 1
        filename = sanitize_filename(f"{self.name}-loopdata.csv")
        filepath = os.path.join(self.outdir, filename)
        df.to_csv(filepath, index_label="idx")

    def writeloops(self) -> None:
        for n, loop in enumerate(self.parent.loops):
            df = loop.df
            filename = sanitize_filename(f"{self.name}-{n+1}.csv")
            filepath = os.path.join(self.outdir, filename)
            df["datetime"] = pd.to_datetime(df.index, unit="ns")
            df.to_csv(
                filepath,
                date_format="%Y-%m-%d %H:%M:%S.%f",
                index_label="timens",
            )


class POIExporter:
    def __init__(self, parent, name) -> None:
        self.parent = parent
        self.name = name
        try:
            self.outdir = parent.properties["dircache"]
        except KeyError:
            self.outdir = os.path.expanduser("~")

    def poi(self) -> None:
        filepath, _ = askSaveFilePath("Export to", f"{self.name}-poi.csv", self.outdir)
        if filepath is None:
            # Cancel pressed
            return
        self.outdir = os.path.dirname(filepath)

        srcseries = self.parent.curve.series
        poiidx = self.parent.curve.feetitem.indices[self.parent.pointkey]
        pois = srcseries[poiidx.dropna()].rename("poi")
        sers = [srcseries, pois]
        df = pd.concat(sers, axis=1, keys=[s.name for s in sers])
        df["datetime"] = pd.to_datetime(df.index, unit="ns")
        df.to_csv(filepath, date_format="%Y-%m-%d %H:%M:%S.%f", index_label="timens")



from typing import List

import numpy as np
import pandas as pd
from physiocurve.pressure.foot import find_tangent_intersections

from graphysio.dialogs import askUserValue
from graphysio.plotwidgets import PlotWidget
from graphysio.plotwidgets.curves import CurveItemWithPOI
from graphysio.structures import Parameter, PlotData


def get_precise_feet(plotwidget: PlotWidget) -> List[CurveItemWithPOI]:
    curvenames = list(plotwidget.curves.keys())
    if len(curvenames) < 1:
        msg = "No curve"
        raise ValueError(msg)
    elif len(curvenames) > 1:
        q = Parameter("Select Curve", curvenames)
        curvename = askUserValue(q)
    else:
        curvename = curvenames[0]

    curve = plotwidget.curves[curvename]
    wave = curve.series

    try:
        idxdia = curve.feetitem.indices["diastole"]
        idxsys = curve.feetitem.indices["systole"]
    except KeyError as e:
        msg = "No systole or diastole"
        raise ValueError(msg) from e

    argdia = wave.index.get_indexer(idxdia)
    argsys = wave.index.get_indexer(idxsys)

    intersections = find_tangent_intersections(wave.to_numpy(), argdia, argsys)

    scale = (wave.index[-1] - wave.index[0]) / len(wave)
    intersections = intersections * scale + wave.index[0]
    intersections = np.rint(intersections).astype(np.int64)

    y_value_index = wave.index.get_indexer(intersections, method="nearest")
    y_values = wave.iloc[y_value_index]

    out_series = pd.Series(y_values.to_numpy(), index=intersections)
    sname = f"{wave.name}-precise_feet"

    plotdata = PlotData(data=out_series, name=sname)
    return [plotdata]



from typing import List

import numpy as np
import pandas as pd
from physiocurve.pressure.foot import find_tangent_intersections
from physiocurve.pressure.incycle import find_dia_sys

from graphysio.dialogs import askUserValue
from graphysio.plotwidgets import PlotWidget
from graphysio.plotwidgets.curves import CurveItemWithPOI
from graphysio.structures import Parameter, PlotData
from graphysio.utils import truncatevecs


def get_tt(plotwidget: PlotWidget) -> List[CurveItemWithPOI]:
    feetitemhash = build_feet_dict(plotwidget.curves)

    param = Parameter("Choose starting points", list(feetitemhash))
    qresult = askUserValue(param)
    beg_curve, beg_feetname = feetitemhash[qresult]

    param = Parameter("Choose ending points", list(feetitemhash))
    qresult = askUserValue(param)
    end_curve, end_feetname = feetitemhash[qresult]

    idxbeg = beg_curve.feetitem.indices[beg_feetname]
    idxend = end_curve.feetitem.indices[end_feetname]
    idxbeg, idxend = truncatevecs([idxbeg, idxend])

    tt = np.abs(idxend - idxbeg)  # nanoseconds
    tt /= 1e6  # milliseconds

    out_series = pd.Series(tt, index=idxbeg)
    sname = f"TT [{beg_curve.name()}_{beg_feetname} - {end_curve.name()}_{end_feetname}] (ms)"
    plotdata = PlotData(data=out_series, name=sname)

    return [plotdata]


def get_pat(plotwidget: PlotWidget) -> List[CurveItemWithPOI]:
    curvenames = list(plotwidget.curves.keys())
    if len(curvenames) < 2:
        msg = "Insufficient data"
        raise ValueError(msg)
    q = Parameter("Select ECG curve", curvenames)
    curvename = askUserValue(q)
    ecgcurve = plotwidget.curves[curvename]

    q = Parameter("Select Pulse curve", curvenames)
    curvename = askUserValue(q)
    pulsecurve = plotwidget.curves[curvename]
    pulse = pulsecurve.series

    try:
        idxfeet = pulsecurve.feetitem.indices["start"]
        idxrwave = ecgcurve.feetitem.indices["rwave"]
    except KeyError as e:
        msg = "Detect feet and R waves first"
        raise ValueError(msg) from e

    argfeet = pulse.index.get_indexer(idxfeet)
    argdia, argsys = find_dia_sys(pulse.to_numpy(), pulsecurve.samplerate, argfeet)
    precise_feet = find_tangent_intersections(pulse.to_numpy(), argdia, argsys)

    time_scale = (pulse.index[-1] - pulse.index[0]) / len(pulse)
    idx_precise_feet = precise_feet * time_scale + pulse.index[0]

    idx_precise_feet, idxrwave = truncatevecs([idx_precise_feet, idxrwave])

    pat = np.abs(idx_precise_feet - idxrwave)  # nanoseconds
    pat /= 1e6  # milliseconds

    out_series = pd.Series(pat, index=idxrwave)
    sname = f"PAT {pulse.name} (ms)"
    plotdata = PlotData(data=out_series, name=sname)

    return [plotdata]


def build_feet_dict(curves):
    feetitemhash = {}
    for curve in curves.values():
        feetitemhash.update(
            {
                f"{curve.name()}-{feetname}": (curve, feetname)
                for feetname in curve.feetitem.indices
            },
        )
    return feetitemhash



from .feet_time_interval import get_feet_time_interval
from .feet_to_curve import get_feet_to_curve
from .perfusion_index import get_perfusion_index
from .precise_feet import get_precise_feet
from .pulse_transit import get_pat, get_tt

Transformations = {
    "Precise Feet": get_precise_feet,
    "Feet to Curve": get_feet_to_curve,
    "Perfusion Index": get_perfusion_index,
    "Pulse Arrival Time": get_pat,
    "Pulse Transit Time": get_tt,
    "Feet Time Interval": get_feet_time_interval,
}



from typing import TYPE_CHECKING

from graphysio.dialogs import askOpenFilePath

from .csv import CsvReader
from .edf import EdfReader
from .parquet import ParquetReader
from .dwc import DwcReader

if TYPE_CHECKING:
    import pathlib

file_readers = {"csv": CsvReader, "parquet": ParquetReader, "edf": EdfReader}
file_readers = {k: mod for k, mod in file_readers.items() if mod.is_available}


class FileReader:
    def __init__(self) -> None:
        super().__init__()
        self.reader = None

        filters = ";;".join(
            [f"{ext.upper()} files (*.{ext})(*.{ext})" for ext in file_readers],
        )
        supported = " ".join(f"*.{ext}" for ext in file_readers)
        self.file_filters = f"All supported ({supported});;{filters}"

    def askFile(self, folder="") -> "pathlib.PurePath":
        filepath, ext = askOpenFilePath(
            "Open File",
            folder=folder,
            filter=self.file_filters,
        )
        if not filepath:
            return folder
        self.reader = file_readers[ext]()
        self.reader.set_data({"filepath": filepath})
        self.reader.askUserInput()
        # Return the parent folder for caching
        return filepath.parent

    # Meant to be executed in seperate thread
    def get_plotdata(self):
        if self.reader:
            return self.reader()
        else:
            return None



from graphysio.plotwidgets.plotwidget import PlotWidget, TimeAxisItem
from graphysio.plotwidgets.poiselector import POISelectorWidget
from graphysio.plotwidgets.puplot import LoopWidget
from graphysio.plotwidgets.spectrogram import SpectrogramWidget
from graphysio.plotwidgets.tsplot import TSWidget

__all__ = [
    "PlotWidget",
    "TimeAxisItem",
    "POISelectorWidget",
    "LoopWidget",
    "SpectrogramWidget",
    "TSWidget",
]



from collections import namedtuple
from enum import Enum
from pathlib import Path
from typing import Optional

import pandas as pd

Filter = namedtuple("Filter", ["name", "parameters"])
Parameter = namedtuple("Parameter", ["description", "request"])


class CycleId(Enum):
    none = "None"
    foot = "Pressure foot"
    pressure = "Pressure Full"
    velocity = "Velocity"
    rwave = "ECG R Wave"
    foottan = "Pressure foot (tangents)"
    pressurebis = "Pressure full (physiocurve)"
    ecg = "ECG full"


class PlotData:
    def __init__(
        self,
        data,
        filepath: Optional[Path] = None,
        name: Optional[str] = None,
    ) -> None:
        if filepath is None and name is None:
            msg = "At least one of filepath or name needs to be specified."
            raise ValueError(msg)
        if filepath is not None:
            self.filepath = Path(filepath)
        self._name = name
        if isinstance(data, pd.Series):
            self.data = data.to_frame(name=self.name)
        else:
            self.data = data

    @property
    def name(self):
        return self._name if self._name is not None else self.filepath.stem

    @property
    def folder(self):
        return self.filepath.parent



import os
import sys
from functools import partial
from itertools import cycle

import numpy as np
import pathvalidate
from pyqtgraph.Qt import QtGui, QtWidgets

sanitize_filename = partial(pathvalidate.sanitize_filename, platform="auto")
sanitize_filepath = partial(pathvalidate.sanitize_filepath, platform="auto")

import importlib.util


def Colors():
    qtcolors = [
        QtGui.QColor(0, 114, 189),
        QtGui.QColor(217, 83, 25),
        QtGui.QColor(237, 177, 32),
        QtGui.QColor(126, 47, 142),
        QtGui.QColor(119, 172, 48),
        QtGui.QColor(77, 190, 238),
        QtGui.QColor(162, 20, 47),
    ]
    return cycle(qtcolors)


def estimateSampleRate(series):
    intervals = np.diff(series.index)
    # 1e9 to account for ns -> Hz
    fs = 1e9 / np.median(intervals)
    if fs == np.inf:
        fs = 0
    if fs > 1:
        fs = int(round(fs))
    return fs


def loadmodule() -> None:
    defaultdir = os.path.expanduser("~")
    filepath = QtWidgets.QFileDialog.getOpenFileName(
        caption="Import module",
        filter="Python files (*.py)",
        directory=defaultdir,
    )
    if not isinstance(filepath, str):
        # PyQt5 API change
        filepath = filepath[0]
    if not filepath:
        # Cancel pressed
        return

    bcbak = sys.dont_write_bytecode
    try:
        sys.dont_write_bytecode = True
        spec = importlib.util.spec_from_file_location("graphysio.plugin", filepath)
        foo = importlib.util.module_from_spec(spec)
        sys.modules["graphysio.plugin"] = foo
        spec.loader.exec_module(foo)
    finally:
        sys.dont_write_bytecode = bcbak


def clip(vec, vrange):
    if vrange is None:
        return vec
    xmin, xmax = vrange
    cond = (vec > xmin) & (vec < xmax)
    return vec[cond]


def truncatevecs(vecs):
    # Ensure all vectors have the same length by truncating the end
    maxidx = min(map(len, vecs))
    return [vec[0:maxidx] for vec in vecs]


def displayError(errmsg) -> None:
    msgbox = QtWidgets.QMessageBox()
    msgbox.setWindowTitle("Error")
    msgbox.setText(str(errmsg))
    msgbox.setStandardButtons(QtWidgets.QMessageBox.Ok)
    msgbox.setIcon(QtWidgets.QMessageBox.Critical)
    msgbox.exec_()
